==> proxy-middleman-0.0.17.tar.gz: proxy-middleman-0.0.17/proxy_middleman/proxy_rotation.py <==
import asyncio
import random
import time
import apprise
from curl_cffi import requests
from curl_cffi.requests.exceptions import ProxyError, ConnectTimeout
from curl_cffi.requests import AsyncSession
from proxy_middleman.custom_exceptions import NoProxiesAvailable, ProxyRetriesExceeded


class ProxyMiddleman:
    BAD_RESPONSE_CODE = {401, 403, 451, 429}
    
    def __init__(
            self, 
            connectors_ids: list[str], 
            slack_webhook_url: str,
            get_proxy_url: str,
            number_of_retries: int = 5, 
            backoff_time: int = 10, 
            exponential_backoff: bool = True,
            use_semaphore: bool = False,
            request_per_proxy: int = 1,
            send_alert_after_unexpected_errors: int = 5,
            bad_response_code: set = None,
        ):
        """
        Initializes the ProxyMiddleman instance with configuration parameters.

        :param connectors_ids: The IDs of the connectors to use.
        :param number_of_retries: Number of retries for proxy requests.
        :param backoff_time: Base time for backoff in seconds.
        :param exponential_backoff: Boolean indicating if exponential backoff should be used.
        :param slack_webhook_url: The URL of the Slack webhook to use for notifications.
        :param use_semaphore: Boolean indicating if semaphore should be used.
        :param request_per_proxy: Number of requests per proxy.
        :param send_alert_after_unexpected_errors: Number of unexpected errors after which an alert should be sent.
        """
        self.connectors = {}  # Save connectors as {connector_id: proxies}
        self.number_of_retries = number_of_retries
        self.backoff_time = backoff_time
        self.exponential_backoff = exponential_backoff
        self.get_proxy_url = get_proxy_url

        if bad_response_code:
            self.BAD_RESPONSE_CODE.update(bad_response_code)

        self.use_semaphore = use_semaphore
        self.request_per_proxy = request_per_proxy
        
        self.send_alert_after_unexpected_errors = send_alert_after_unexpected_errors

        self.semaphores = {}
        for connector_id in connectors_ids:
            self.connectors[connector_id], semaphore_value = self._initialize_proxies_and_semaphore(connector_id)
            if semaphore_value:
                self.semaphores[connector_id] = asyncio.Semaphore(semaphore_value)

        self.apprise_obj = apprise.Apprise()
        self.apprise_obj.add(slack_webhook_url)

    def _initialize_proxies_and_semaphore(self, connector_id):
        """
        Requests and initializes a list of proxies for a specific connector_id.

        :param connector_id: The ID of the connector for which to initialize proxies.
        :return: A dictionary of proxies with their metadata, filtered by those with status "STARTED".
        :raises Exception: If no proxies are found for the given connector_id.
        """
        response = requests.request(method="GET", url=self.get_proxy_url)
        for data in response.json():
            if data["connector_id"] == connector_id:
                proxies = data["proxies"]
                break
        else:
            raise NoProxiesAvailable(connector_id)

        return {
            proxy: {
                "unexpected_fails": 0,
                "proxy_not_working_fails": 0,
                "disabled_until": None,
                "added_at": time.time(),
                "request_count": 0,
            }
            for proxy in proxies
        }, len(proxies) * self.request_per_proxy if self.use_semaphore else None

    async def _get_available_proxies(self, connector_id=None):
        current_time = time.time()
        available_proxies = {}

        connectors = (
            {connector_id: self.connectors[connector_id]}
            if connector_id and connector_id in self.connectors
            else self.connectors
        )

        for conn_id, proxies in connectors.items():
            filtered_proxies = [
                p for p, meta in proxies.items()
                if meta["disabled_until"] is None or meta["disabled_until"] < current_time
            ]
            if filtered_proxies:
                available_proxies[conn_id] = filtered_proxies

        return available_proxies

    async def _get_proxy(self):
        last_notification_time = 0

        while True:
            available_proxies = await self._get_available_proxies()

            if available_proxies:
                connector_id, proxies = random.choice(list(available_proxies.items()))
                return connector_id, random.choice(proxies), self.semaphores.get(connector_id)

            current_time = time.time()
            if current_time - last_notification_time >= 300:  # 5 minutes
                self.notify("No available proxies! Sleeping for 5 minutes...", notify_type=apprise.NotifyType.WARNING)
                last_notification_time = current_time

            await asyncio.sleep(30)

    async def _disable_proxy(self, connector_id, proxy, unexpected_error=False, proxy_not_working=False, response_text=None):
        """
        Disables a proxy for a given connector_id. If a proxy fails 3 times in a row, it gets removed.

        :param connector_id: The ID of the connector containing the proxy.
        :param proxy: The name of the proxy to disable.
        """
        if connector_id in self.connectors and proxy in self.connectors[connector_id]:
            if unexpected_error:
                self.connectors[connector_id][proxy]["unexpected_fails"] += 1
            elif proxy_not_working:
                self.connectors[connector_id][proxy]["proxy_not_working_fails"] += 1
            else:
                raise ValueError("unexpected_error or proxy_not_working must be True")

            # If the proxy breaks 3 times in a row, delete it
            if self.connectors[connector_id][proxy]["proxy_not_working_fails"] >= 3:
                self.notify(f"Proxy '{proxy}' removed after 3 consecutive failures. Response text: {response_text}", notify_type=apprise.NotifyType.WARNING)
                del self.connectors[connector_id][proxy]
                await self._recalculate_semaphore(connector_id)
            else:
                # Normal shutdown with backoff
                max_fails = max(self.connectors[connector_id][proxy]["unexpected_fails"], self.connectors[connector_id][proxy]["proxy_not_working_fails"])
                backoff = self.backoff_time * (2 ** max_fails) if self.exponential_backoff else self.backoff_time
                self.connectors[connector_id][proxy]["disabled_until"] = time.time() + backoff

            total_unexpected_fails = sum(proxy_data["unexpected_fails"] for proxy_data in self.connectors[connector_id].values())

            if total_unexpected_fails % self.send_alert_after_unexpected_errors == 0:
                self.notify(f"There have been {self.send_alert_after_unexpected_errors} unexpected errors. Response text: {response_text}", notify_type=apprise.NotifyType.WARNING)
    
    async def _get_semaphore(self):
        """
        :return: A semaphore for the availble_proxies_count * request_per_proxy.
        """
        _, _, semaphore = await self._get_proxy()
        return semaphore
    
    async def _recalculate_semaphore(self, connector_id: str):
        if not self.use_semaphore:
            return None

        all_proxies = self.connectors.get(connector_id, {})
        semaphore_value = len(all_proxies) * self.request_per_proxy

        if semaphore_value > 0:
            self.semaphores[connector_id] = asyncio.Semaphore(semaphore_value)
        else:
            self.notify(f"No proxies available for connector {connector_id}.", notify_type=apprise.NotifyType.WARNING)

    async def _request(self, url, method="GET", **kwargs):
        """
        Sends an HTTP request using a proxy, with retries on failure.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (default is "GET").
        :param kwargs: Additional arguments to pass to the request.
        :return: The HTTP response object.
        :raises Exception: If all proxies fail after the specified number of retries.
        """
        for _ in range(self.number_of_retries):
            connector_id, proxy, _ = await self._get_proxy()
            try:
                async with AsyncSession() as session:
                    response = await session.request(method, url, proxy=proxy, **kwargs)
                
                self.connectors[connector_id][proxy]["request_count"] += 1
                
                if response.status_code in self.BAD_RESPONSE_CODE:
                    await self._disable_proxy(connector_id, proxy, unexpected_error=True, response_text=response.text)
                    continue
                self.connectors[connector_id][proxy]["proxy_not_working_fails"] = 0
                return response
            except UnicodeDecodeError as e:
                self.notify(f"Unicode error: {e}", notify_type=apprise.NotifyType.FAILURE)
                return None
            except ConnectTimeout:
                pass
            except ProxyError:
                await self._disable_proxy(connector_id, proxy, proxy_not_working=True, response_text=response.text)
            except Exception as e:
                await self._disable_proxy(connector_id, proxy, unexpected_error=True, response_text=e)
        
        self.notify(f"{self.number_of_retries} attempts were made to send a request using a proxy and none of them passed. Raising an exception on the backend", notify_type=apprise.NotifyType.FAILURE)
        # raise ProxyRetriesExceeded(self.number_of_retries) 

    async def request(self, url, method="GET", **kwargs):
        """
        Sends an HTTP request using a proxy, with retries on failure.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (default is "GET").
        :param kwargs: Additional arguments to pass to the request.
        :return: The HTTP response object.
        :raises Exception: If all proxies fail after the specified number of retries.
        """

        if self.use_semaphore:
            semaphore = await self._get_semaphore()
            async with semaphore:
                return await self._request(url, method, **kwargs)
        else:
            return await self._request(url, method, **kwargs)

    def notify(self, message, notify_type=apprise.NotifyType.INFO):
        """
        Sends a notification with the given message.

        :param message: The message to include in the notification.
        :param notify_type: The type of the notification.
        """
        try:
            if self.apprise_obj:
                self.apprise_obj.notify(
                    body=message,
                    title="Middleman Proxy Info",
                    notify_type=notify_type
                )
            else:
                print("No apprise object found. Skipping notification")
        except Exception as e:
            print(f"Error sending notification: {e}")

    def get_proxy_stats(self):
        """
        Returns statistics for all proxies in all connectors.

        :return: A dictionary containing statistics for each proxy, including request count, lifetime, and fails.
        """
        return {
            connector_id: {
                proxy: {
                        "request_count": meta["request_count"],
                        "lifetime_minutes": round((time.time() - meta["added_at"]) / 60, 2),
                        "unexpected_fails": meta["unexpected_fails"],
                        "proxy_not_working_fails": meta["proxy_not_working_fails"],
                        "disabled_until": meta["disabled_until"]
                    }
                    for proxy, meta in proxies.items()
                }
                for connector_id, proxies in self.connectors.items()
            }

==> proxy-middleman-0.0.18.tar.gz: proxy-middleman-0.0.18/proxy_middleman/proxy_rotation.py <==
import asyncio
import random
import textwrap
import time
import apprise
from curl_cffi import requests
from curl_cffi.requests.exceptions import ProxyError, ConnectTimeout
from curl_cffi.requests import AsyncSession
from proxy_middleman.custom_exceptions import NoProxiesAvailable #, ProxyRetriesExceeded
from proxy_middleman.notification import SlackNotifier


class ProxyMiddleman:
    BAD_RESPONSE_CODE = {401, 403, 451, 429}
    
    def __init__(
            self, 
            connectors_ids: list[str], 
            slack_webhook_url: str,
            get_proxy_url: str,
            number_of_retries: int = 5, 
            backoff_time: int = 10, 
            exponential_backoff: bool = True,
            use_semaphore: bool = False,
            request_per_proxy: int = 1,
            send_alert_after_unexpected_errors: int = 5,
            bad_response_code: set = None,
            notifier: SlackNotifier = None,
        ):
        """
        Initializes the ProxyMiddleman instance with configuration parameters.

        :param connectors_ids: The IDs of the connectors to use.
        :param number_of_retries: Number of retries for proxy requests.
        :param backoff_time: Base time for backoff in seconds.
        :param exponential_backoff: Boolean indicating if exponential backoff should be used.
        :param slack_webhook_url: The URL of the Slack webhook to use for notifications.
        :param use_semaphore: Boolean indicating if semaphore should be used.
        :param request_per_proxy: Number of requests per proxy.
        :param send_alert_after_unexpected_errors: Number of unexpected errors after which an alert should be sent.
        :param get_proxy_url: Endpoint url
        :param notifier: SlackNotifier class instance
        """
        self.connectors = {}  # Save connectors as {connector_id: proxies}
        self.number_of_retries = number_of_retries
        self.backoff_time = backoff_time
        self.exponential_backoff = exponential_backoff
        self.get_proxy_url = get_proxy_url

        if bad_response_code:
            self.BAD_RESPONSE_CODE.update(bad_response_code)

        self.use_semaphore = use_semaphore
        self.request_per_proxy = request_per_proxy
        
        self.send_alert_after_unexpected_errors = send_alert_after_unexpected_errors

        self.semaphores = {}
        for connector_id in connectors_ids:
            self.connectors[connector_id], semaphore_value = self._initialize_proxies_and_semaphore(connector_id)
            if semaphore_value:
                self.semaphores[connector_id] = asyncio.Semaphore(semaphore_value)

        self.notifier = notifier if notifier else SlackNotifier(slack_webhook_url)
        
    def _initialize_proxies_and_semaphore(self, connector_id):
        """
        Requests and initializes a list of proxies for a specific connector_id.

        :param connector_id: The ID of the connector for which to initialize proxies.
        :return: A dictionary of proxies with their metadata, filtered by those with status "STARTED".
        :raises Exception: If no proxies are found for the given connector_id.
        """
        response = requests.request(method="GET", url=self.get_proxy_url)
        for data in response.json():
            if data["connector_id"] == connector_id:
                proxies = data["proxies"]
                break
        else:
            raise NoProxiesAvailable(connector_id)

        return {
            proxy: {
                "unexpected_fails": 0,
                "proxy_not_working_fails": 0,
                "disabled_until": None,
                "added_at": time.time(),
                "request_count": 0,
            }
            for proxy in proxies
        }, len(proxies) * self.request_per_proxy if self.use_semaphore else None

    async def _get_available_proxies(self, connector_id=None):
        current_time = time.time()
        available_proxies = {}

        connectors = (
            {connector_id: self.connectors[connector_id]}
            if connector_id and connector_id in self.connectors
            else self.connectors
        )

        for conn_id, proxies in connectors.items():
            filtered_proxies = [
                p for p, meta in proxies.items()
                if meta["disabled_until"] is None or meta["disabled_until"] < current_time
            ]
            if filtered_proxies:
                available_proxies[conn_id] = filtered_proxies

        return available_proxies

    async def _get_proxy(self):
        last_notification_time = 0

        while True:
            available_proxies = await self._get_available_proxies()

            if available_proxies:
                connector_id, proxies = random.choice(list(available_proxies.items()))
                return connector_id, random.choice(proxies), self.semaphores.get(connector_id)

            current_time = time.time()
            if current_time - last_notification_time >= 300:  # 5 minutes
                self.notifier.notify("No available proxies! Sleeping for 5 minutes...", title="Middleman Proxy Info", notify_type=apprise.NotifyType.WARNING)
                last_notification_time = current_time

            await asyncio.sleep(30)

    async def _disable_proxy(self, connector_id, proxy, unexpected_error=False, proxy_not_working=False, response_text=None):
        """
        Disables a proxy for a given connector_id. If a proxy fails 3 times in a row, it gets removed.

        :param connector_id: The ID of the connector containing the proxy.
        :param proxy: The name of the proxy to disable.
        """
        if connector_id in self.connectors and proxy in self.connectors[connector_id]:
            if unexpected_error:
                self.connectors[connector_id][proxy]["unexpected_fails"] += 1
            elif proxy_not_working:
                self.connectors[connector_id][proxy]["proxy_not_working_fails"] += 1
            else:
                raise ValueError("unexpected_error or proxy_not_working must be True")

            # If the proxy breaks 3 times in a row, delete it
            if self.connectors[connector_id][proxy]["proxy_not_working_fails"] >= 3:
                self.notifier.notify(f"Proxy '{proxy}' removed after 3 consecutive failures. Response text: {response_text}", title="Middleman Proxy Info", notify_type=apprise.NotifyType.WARNING)
                del self.connectors[connector_id][proxy]
                await self._recalculate_semaphore(connector_id)
            else:
                # Normal shutdown with backoff
                max_fails = max(self.connectors[connector_id][proxy]["unexpected_fails"], self.connectors[connector_id][proxy]["proxy_not_working_fails"])
                backoff = self.backoff_time * (2 ** max_fails) if self.exponential_backoff else self.backoff_time
                self.connectors[connector_id][proxy]["disabled_until"] = time.time() + backoff
 
    async def _get_semaphore(self):
        """
        :return: A semaphore for the availble_proxies_count * request_per_proxy.
        """
        _, _, semaphore = await self._get_proxy()
        return semaphore
    
    async def _recalculate_semaphore(self, connector_id: str):
        if not self.use_semaphore:
            return None

        all_proxies = self.connectors.get(connector_id, {})
        semaphore_value = len(all_proxies) * self.request_per_proxy

        if semaphore_value > 0:
            self.semaphores[connector_id] = asyncio.Semaphore(semaphore_value)
        else:
            self.notifier.notify(f"No proxies available for connector {connector_id}.", title="Middleman Proxy Info", notify_type=apprise.NotifyType.WARNING)

    async def _request(self, url, method="GET", **kwargs):
        """
        Sends an HTTP request using a proxy, with retries on failure.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (default is "GET").
        :param kwargs: Additional arguments to pass to the request.
        :return: The HTTP response object.
        :raises Exception: If all proxies fail after the specified number of retries.
        """
        last_error = None

        for _ in range(self.number_of_retries):
            connector_id, proxy, _ = await self._get_proxy()
            try:
                async with AsyncSession() as session:
                    response = await session.request(method, url, proxy=proxy, **kwargs)
                
                self.connectors[connector_id][proxy]["request_count"] += 1
                
                if response.status_code in self.BAD_RESPONSE_CODE:
                    await self._disable_proxy(connector_id, proxy, unexpected_error=True, response_text=response.text)
                    continue
                self.connectors[connector_id][proxy]["proxy_not_working_fails"] = 0
                return response
            except UnicodeDecodeError as unicode_error:
                last_error = unicode_error
                break
            except ConnectTimeout:
                pass
            except ProxyError as proxy_error:
                last_error = proxy_error
                await self._disable_proxy(connector_id, proxy, proxy_not_working=True, response_text=response.text)
            except Exception as unexpected_error:
                last_error = unexpected_error
                await self._disable_proxy(connector_id, proxy, unexpected_error=True)

        shortened_kwargs = {
            key: textwrap.shorten(str(value), width=30, placeholder="...")
            for key, value in kwargs.items()
        }
        error = (
            f"Url - {url}\n"
            f"Method - {method}\n"
            f"Kwargs - {shortened_kwargs}\n"
            f"Type - {type(last_error)}\n"
            f"Msg - {last_error}"
        )
        self.notifier.notify(error, title="Middleman Proxy Info", notify_type=apprise.NotifyType.FAILURE)
        # raise ProxyRetriesExceeded(self.number_of_retries) 

    async def request(self, url, method="GET", **kwargs):
        """
        Sends an HTTP request using a proxy, with retries on failure.

        :param url: The URL to send the request to.
        :param method: The HTTP method to use (default is "GET").
        :param kwargs: Additional arguments to pass to the request.
        :return: The HTTP response object.
        :raises Exception: If all proxies fail after the specified number of retries.
        """

        if self.use_semaphore:
            semaphore = await self._get_semaphore()
            async with semaphore:
                return await self._request(url, method, **kwargs)
        else:
            return await self._request(url, method, **kwargs)

    def notify(self, message, notify_type=apprise.NotifyType.INFO):
        """
        Sends a notification with the given message.

        :param message: The message to include in the notification.
        :param notify_type: The type of the notification.
        """
        try:
            if self.apprise_obj:
                self.apprise_obj.notify(
                    body=message,
                    title="Middleman Proxy Info",
                    notify_type=notify_type
                )
            else:
                print("No apprise object found. Skipping notification")
        except Exception as e:
            print(f"Error sending notification: {e}")

    def get_proxy_stats(self):
        """
        Returns statistics for all proxies in all connectors.

        :return: A dictionary containing statistics for each proxy, including request count, lifetime, and fails.
        """
        return {
            connector_id: {
                proxy: {
                        "request_count": meta["request_count"],
                        "lifetime_minutes": round((time.time() - meta["added_at"]) / 60, 2),
                        "unexpected_fails": meta["unexpected_fails"],
                        "proxy_not_working_fails": meta["proxy_not_working_fails"],
                        "disabled_until": meta["disabled_until"]
                    }
                    for proxy, meta in proxies.items()
                }
                for connector_id, proxies in self.connectors.items()
            }
